/*
 * Copyright 2006 Google Inc.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.google.gwt.user.rebind.rpc;

import com.google.gwt.core.ext.GeneratorContext;
import com.google.gwt.core.ext.TreeLogger;
import com.google.gwt.core.ext.typeinfo.JArrayType;
import com.google.gwt.core.ext.typeinfo.JClassType;
import com.google.gwt.core.ext.typeinfo.JField;
import com.google.gwt.core.ext.typeinfo.JType;
import com.google.gwt.user.client.rpc.SerializationException;
import com.google.gwt.user.client.rpc.SerializationStreamReader;
import com.google.gwt.user.client.rpc.SerializationStreamWriter;
import com.google.gwt.user.rebind.ClassSourceFileComposerFactory;
import com.google.gwt.user.rebind.SourceWriter;

import java.io.PrintWriter;

/**
 * <p>
 * This class creates field serializers for a particular class. If the class has
 * a custom serializer then that class is used rather than creating one.
 * 
 * <p>
 * Generated field serializers are emitted into the same package as the class
 * that they serialize.
 * 
 * <p>
 * Fields are considered serializable if:
 * <ul>
 * <li>Field is not static
 * <li>Field is not transient
 * </ul>
 * 
 * TODO(mmendez): Need to make the generated field serializers final
 * TODO(mmendez): Would be nice to be able to have imports, rather than using
 * fully qualified type names everywhere
 */
public class FieldSerializerCreator {
  public FieldSerializerCreator(SerializableTypeOracle serializationOracle,
      JType requestedClass) {
    setSerializationOracle(serializationOracle);
    fSerializableClass = requestedClass;
  }

  public String realize(TreeLogger logger, GeneratorContext ctx) {
    assert (ctx != null);
    assert (fSerializationOracle.isSerializable(getSerializableClass()));

    logger = logger.branch(TreeLogger.SPAM,
      "Generating a field serializer for type '"
        + fSerializableClass.getQualifiedSourceName() + "'", null);
    String fieldSerializerName = fSerializationOracle.getFieldSerializerName(getSerializableClass());

    fSourceWriter = getSourceWriter(logger, ctx);
    if (fSourceWriter == null) {
      return fieldSerializerName;
    }
    assert fSourceWriter != null;

    writeFieldAccessors();

    writeSerializeMethod();

    writeDeserializeMethod();

    fSourceWriter.commit(logger);

    return fieldSerializerName;
  }

  /**
   * Return an array of fields declared by this class that are serializable
   */
  private JField[] applyFieldSerializationPolicy() {
    if (fSerializableFields != null) {
      return fSerializableFields;
    }

    fSerializableFields = fSerializationOracle.applyFieldSerializationPolicy(getSerializableClass());
    return fSerializableFields;
  }

  private JType getSerializableClass() {
    return fSerializableClass;
  }

  private String getSerializableClassPackageName() {
    JType type = getSerializableClass();
    String packageName = fSerializationOracle.getFieldSerializerName(type);
    int packageNameEnd = packageName.lastIndexOf('.');
    if (packageNameEnd >= 0) {
      packageName = packageName.substring(0, packageNameEnd);
    }

    return packageName;
  }

  /**
   * Given a class return the next superclass that is serializable.
   * 
   * @param serializableClass
   * @return
   */
  private JClassType getSerializableSuperclass(JClassType serializableClass) {
    if (serializableClass == null) {
      return null;
    }

    JClassType superClass = serializableClass.getSuperclass();
    if (superClass != null) {
      SerializableTypeOracle serializationOracle = getSerializationOracle();
      if (serializationOracle.isSerializable(superClass)) {
        return superClass;
      }

      return getSerializableSuperclass(superClass);
    }

    return null;
  }

  private SerializableTypeOracle getSerializationOracle() {
    return fSerializationOracle;
  }

  private String getSimpleSourceName(JType type) {
    String sourceName = fSerializationOracle.getFieldSerializerName(type);

    int qualifiedSourceNameStart = sourceName.lastIndexOf('.');
    if (qualifiedSourceNameStart >= 0) {
      sourceName = sourceName.substring(qualifiedSourceNameStart + 1);
    }

    return sourceName;
  }

  private SourceWriter getSourceWriter(TreeLogger logger, GeneratorContext ctx) {
    String packageName = getSerializableClassPackageName();
    String className = getSimpleSourceName(getSerializableClass());

    PrintWriter printWriter = ctx.tryCreate(logger, packageName, className);
    if (printWriter == null) {
      return null;
    }

    ClassSourceFileComposerFactory composerFactory = new ClassSourceFileComposerFactory(
      packageName, className);

    return composerFactory.createSourceWriter(ctx, printWriter);
  }

  /**
   * Returns true if we will need get/set method pair for a field. These are
   * required when the field would not be accessible to the serializer class.
   */
  private boolean needsAccessorMethods(JField field) {
    boolean needsAccessors = !field.isPublic();
    return needsAccessors;
  }

  private void setSerializationOracle(SerializableTypeOracle serializationOracle) {
    fSerializationOracle = serializationOracle;
  }

  private void writeDeserializeMethod() {
    fSourceWriter.print("public static void deserialize(");
    fSourceWriter.print(SerializationStreamReader.class.getName());
    fSourceWriter.print(" streamReader, ");
    fSourceWriter.print(getSerializableClass().getQualifiedSourceName());
    fSourceWriter.println(" instance) throws "
      + SerializationException.class.getName() + "{");
    fSourceWriter.indent();

    writeFieldDeserializationStatements();

    JClassType serializableClass = getSerializableClass().isClass();
    if (serializableClass != null) {
      JClassType serializableSuperClass = getSerializableSuperclass(serializableClass);
      if (serializableSuperClass != null) {
        String fieldSerializerName = fSerializationOracle.getFieldSerializerName(serializableSuperClass);
        fSourceWriter.print(fieldSerializerName);
        fSourceWriter.println(".deserialize(streamReader, instance);");
      }
    }

    fSourceWriter.outdent();
    fSourceWriter.println("}");
    fSourceWriter.println();
  }

  /**
   * This method will generate a native JSNI accessor method for every field
   * that is protected, private using the "Violator" pattern to allow an
   * external class to access the field's value.
   */
  private void writeFieldAccessors() {
    JField[] serializableFields = applyFieldSerializationPolicy();
    int fieldCount = serializableFields.length;

    for (int fieldIndex = 0; fieldIndex < fieldCount; ++fieldIndex) {
      JField serializableField = serializableFields[fieldIndex];

      if (!needsAccessorMethods(serializableField)) {
        continue;
      }

      writeFieldGet(serializableField);
      writeFieldSet(serializableField);
    }
  }

  private void writeFieldDeserializationStatements() {
    JType type = getSerializableClass();
    JArrayType isArray = type.isArray();
    if (isArray != null) {
      fSourceWriter.println("for (int itemIndex = 0; itemIndex < instance.length; ++itemIndex) {");
      fSourceWriter.indent();

      JType componentType = isArray.getComponentType();

      fSourceWriter.print("instance[itemIndex] = ");
      if (Shared.typeNeedsCast(componentType)) {
        fSourceWriter.print("(" + componentType.getQualifiedSourceName() + ") ");
      }
      String readMethodName = "streamReader.read"
        + Shared.getCallSuffix(componentType);
      fSourceWriter.println(readMethodName + "();");
      fSourceWriter.outdent();
      fSourceWriter.println("}");
    } else {
      JField[] serializableFields = applyFieldSerializationPolicy();
      int fieldCount = serializableFields.length;

      for (int fieldIndex = 0; fieldIndex < fieldCount; ++fieldIndex) {
        JField serializableField = serializableFields[fieldIndex];
        JType fieldType = serializableField.getType();
        boolean needsAccessor = needsAccessorMethods(serializableField);

        String readMethodName = "read" + Shared.getCallSuffix(fieldType);
        String streamReadExpression = "streamReader." + readMethodName + "()";
        if (Shared.typeNeedsCast(fieldType)) {
          streamReadExpression = "(" + fieldType.getQualifiedSourceName()
            + ") " + streamReadExpression;
        }

        if (needsAccessor) {
          fSourceWriter.print("set");
          fSourceWriter.print(Shared.capitalize(serializableField.getName()));
          fSourceWriter.print("(instance, ");
          fSourceWriter.print(streamReadExpression);
          fSourceWriter.println(");");
        } else {
          fSourceWriter.print("instance.");
          fSourceWriter.print(serializableField.getName());
          fSourceWriter.print(" = ");
          fSourceWriter.print(streamReadExpression);
          fSourceWriter.println(";");
        }
      }
    }

    fSourceWriter.println();
  }

  /**
   * Write a getter method for an instance field.
   */
  private void writeFieldGet(JField serializableField) {
    JType fieldType = serializableField.getType();
    String fieldTypeQualifiedSourceName = fieldType.getQualifiedSourceName();
    String serializableClassQualifedName = getSerializableClass().getQualifiedSourceName();
    String fieldName = serializableField.getName();

    fSourceWriter.print("private static native ");
    fSourceWriter.print(fieldTypeQualifiedSourceName);
    fSourceWriter.print(" get");
    fSourceWriter.print(Shared.capitalize(fieldName));
    fSourceWriter.print("(");
    fSourceWriter.print(getSerializableClass().getQualifiedSourceName());
    fSourceWriter.println(" instance) /*-{");
    fSourceWriter.indent();

    fSourceWriter.print("return instance.@");
    fSourceWriter.print(serializableClassQualifedName);
    fSourceWriter.print("::");
    fSourceWriter.print(fieldName);
    fSourceWriter.println(";");

    fSourceWriter.outdent();
    fSourceWriter.println("}-*/;");
    fSourceWriter.println();
  }

  private void writeFieldSerializationStatements() {
    JType type = getSerializableClass();
    JArrayType isArray = type.isArray();
    if (isArray != null) {
      fSourceWriter.println("int itemCount = instance.length;");
      fSourceWriter.println();
      fSourceWriter.println("streamWriter.writeInt(itemCount);");
      fSourceWriter.println();
      fSourceWriter.println("for (int itemIndex = 0; itemIndex < itemCount; ++itemIndex) {");
      fSourceWriter.indent();
      String writeMethodName = "write"
        + Shared.getCallSuffix(isArray.getComponentType());
      fSourceWriter.print("streamWriter.");
      fSourceWriter.print(writeMethodName);
      fSourceWriter.println("(instance[itemIndex]);");
      fSourceWriter.outdent();
      fSourceWriter.println("}");
    } else {
      JField[] serializableFields = applyFieldSerializationPolicy();
      int fieldCount = serializableFields.length;

      for (int fieldIndex = 0; fieldIndex < fieldCount; ++fieldIndex) {
        JField serializableField = serializableFields[fieldIndex];
        JType fieldType = serializableField.getType();

        String writeMethodName = "write" + Shared.getCallSuffix(fieldType);
        fSourceWriter.print("streamWriter.");
        fSourceWriter.print(writeMethodName);
        fSourceWriter.print("(");

        if (needsAccessorMethods(serializableField)) {
          fSourceWriter.print("get");
          fSourceWriter.print(Shared.capitalize(serializableField.getName()));
          fSourceWriter.println("(instance));");
        } else {
          fSourceWriter.print("instance.");
          fSourceWriter.print(serializableField.getName());
          fSourceWriter.println(");");
        }
      }
    }

    fSourceWriter.println();
  }

  /**
   * Write a setter method for an instance field.
   */
  private void writeFieldSet(JField serializableField) {
    JType fieldType = serializableField.getType();
    String fieldTypeQualifiedSourceName = fieldType.getQualifiedSourceName();
    String serializableClassQualifedName = getSerializableClass().getQualifiedSourceName();
    String fieldName = serializableField.getName();

    fSourceWriter.print("private static native void ");
    fSourceWriter.print(" set");
    fSourceWriter.print(Shared.capitalize(fieldName));
    fSourceWriter.print("(");
    fSourceWriter.print(serializableClassQualifedName);
    fSourceWriter.print(" instance, ");
    fSourceWriter.print(fieldTypeQualifiedSourceName);
    fSourceWriter.println(" value) /*-{");
    fSourceWriter.indent();

    fSourceWriter.print("instance.@");
    fSourceWriter.print(serializableClassQualifedName);
    fSourceWriter.print("::");
    fSourceWriter.print(fieldName);
    fSourceWriter.println(" = value;");

    fSourceWriter.outdent();
    fSourceWriter.println("}-*/;");
    fSourceWriter.println();
  }

  private void writeSerializeMethod() {
    fSourceWriter.print("public static void serialize(");
    fSourceWriter.print(SerializationStreamWriter.class.getName());
    fSourceWriter.print(" streamWriter, ");
    fSourceWriter.print(getSerializableClass().getQualifiedSourceName());
    fSourceWriter.println(" instance) throws "
      + SerializationException.class.getName() + " {");
    fSourceWriter.indent();

    writeFieldSerializationStatements();

    JClassType serializableClass = getSerializableClass().isClass();
    if (serializableClass != null) {
      JClassType serializableSuperclass = getSerializableSuperclass(serializableClass);
      if (serializableSuperclass != null) {
        String fieldSerializerName = fSerializationOracle.getFieldSerializerName(serializableSuperclass);
        fSourceWriter.print(fieldSerializerName);
        fSourceWriter.println(".serialize(streamWriter, instance);");
      }
    }

    fSourceWriter.outdent();
    fSourceWriter.println("}");
    fSourceWriter.println();
  }

  private JType fSerializableClass;
  private JField[] fSerializableFields;
  private SerializableTypeOracle fSerializationOracle;
  private SourceWriter fSourceWriter;

}
